;
; Silly lambda manipulations
;
; Copyright (c) 1998 Far Rideau <fare@tunes.org> and the TUNES project.
; Freely distributable under the GNU LGPL v2.1 or later.
;
; This file is independent from the other ones,
; apart from its sharing basic run-time and library services.
; It's all about describing simple aspects of lambda calculi.
; Hopefully, I'll eventually tackle illative combinatory logic in here.
;

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
; importing easy stuff from the Tunes scheme library

(require 'lib)
; imported symbols: foldl1 make-symbol add1 sub1

(require 'macrolib)
; imported macros: trace-on trace-off

;;; default for the above:
;(define trace-on list)(define trace-off list)
;(define (add1 x) (+ x 1)) (define (sub1 x) (- x 1))
;(define (foldr1 f l) (if(null?(cdr l))(car l)(f(car l)(foldr1 f(cdr l)))))
;(define (foldl1 f l) (foldr1 (lambda (x y) (f y x)) (reverse l)))
;(define (*->string x) (cond((string? x)x)((symbol? x)(symbol->string x))
; ((number? x)(number->string x))((char? x)(make-string 1 x))(else "")))
;(define (multi->string . args) (apply string-append (map *->string args)))
;(define (make-symbol . args) (string->symbol (apply multi->string args)))
;(define (trace-on . x) #f)
;(define (trace-off . x) #f)

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
; helpers to define a new structure

(define (make-maker tag)
  (lambda l (apply vector tag l)))
(define (make-detector tag)
  (lambda (x)
    (and (vector? x)
	 (> (vector-length x) 0)
	 (eq? tag (vector-ref x 0)))))
(define (make-accessor n)
  (lambda (x)
    (and (vector-ref x n))))

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
; new structures and helper functions

(define (make-app x y) (list x y))
(define (app? x) (pair? x))
(define (app-fun x) (car x))
(define (app-arg x) (cadr x))

(define (make-lambda x y) ((make-maker 'lambda) x y))
(define lambda? (make-detector 'lambda))
(define lambda-param (make-accessor 1))
(define lambda-body (make-accessor 2))

(define (make-subst x v E) ((make-maker 'subst) x v E))
(define subst? (make-detector 'subst))
(define subst-var (make-accessor 1))
(define subst-val (make-accessor 2))
(define subst-body (make-accessor 3))

(define (foldapp l)
  (foldl1 make-app l))
(define (unfoldapp e)
  (define (u e l)
    (if (app? e) (u (app-fun e) (cons (app-arg e) l))
	(cons e l)))
  (u e '()))

(define (free? x e)
  (cond	((lambda? e) (or (eq? x (lambda-param e)) (free? x (lambda-body e))))
	((subst? e) (or (eq? x (subst-var e))
			(and (free? x (subst-val e))
			     (free? x (subst-body e)))))
	((app? e) (and (free? x (app-fun e)) (free? x (app-arg e))))
	((eq? x e) #f)
	(else #t)))

(define (alpha-l-equal? l M N)
  (cond ((lambda? M)
	 (and (lambda? N)
	      (alpha-l-equal?
	       (cons (cons (lambda-param M) (lambda-param N)) l)
	       (lambda-body M)
	       (lambda-body N))))
	((app? M)
	 (and (app? N)
	      (and (alpha-l-equal? l (app-fun M) (app-fun N))
		   (alpha-l-equal? l (app-arg M) (app-arg N)))))
	((assoc M l)
	 (eq? N (cdr (assoc M l)))) ; not remotely optimized
	(else ; unbound atom
	 (eq? M N))))
(define (alpha-equal? M N) (alpha-l-equal? '() M N))


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
; easier syntax for our little lambda language (w/o subst)
(define (parse e)
  (if (and (pair? e) (list? e))
      (case (car e)
	((lambda) (make-lambda (cadr e) (parse (caddr e))))
	(else (foldapp (map parse e))))
      e))

(define (unparse e)
  (define (unp-lam a)
    (list 'lambda (lambda-param a) (unparse (lambda-body a))))
  (define (appfold a l)
    (if (app? a) (appfold (app-fun a) (cons (unparse (app-arg a)) l))
	(cons (unparse a) l)))
  (cond ((app? e) (appfold e '()))
	((lambda? e) (unp-lam e))
	(else e)))



;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
"
same lambda expressions, but with deBruijn indices instead of variables.

deBruijn indices allow to simply express a terminating and confluent
substitution calculus. The grammar is:
Index ::= '0' | '^' Index
	| 'shiftindex' Index Index
Term ::=  'app' Term Term | 'var' Index | 'lam' Term | 'atom' Atom
	| 'shift' Index Term
	| 'subst' Term Index Term | 'substvar' Term Index Index Index

Constructors shifindex, shift, subst, substvar are used for the sake
of substitution; they are fully eliminated by the rewrite rules.
They are herein considered as part of the calculus, in as much as
the calculus precisely tries to intern substitution as a simple set
of rewrite rules; they might alternatively be considered as functions
external to the grammar of terms proper.
subst N I M is substituting N for the ith variable in M;
variables in a nearer scope are untouched; variables in an outer scope
are made one step nearer to account for a variable layer disappearing;
substitution commutes with application; when going through a lambda,
it has to shift all variable references in N, to account for variables
having to cross one more lexical layer.

Indices are really natural integers, as constructed
by initial number 0 and a successor function ^.
Constructor var is used to denote the variable bound by the ith lambda
above current expression.
The atom constructor is optional, and allows for the calculus to operate
on external objects. Such objects are not affected by substitutions,
but may have some optional effect on computations.

rewrite rules are:
subst N I app A B --> app subst N I A subst N I B
subst N I atom A --> atom A
subst N 0 var 0 --> N
subst N ^I var 0 --> var 0
subst N I var ^K --> substvar N I ^K K
substvar N 0 0 K --> N
substvar N 0 ^J K --> K
substvar N ^I 0 K --> ^K
substvar N ^I ^J K --> substvar N I J K
subst N I lam M --> lam subst shift 0 N ^I M
shift I N app A B --> app shift I N A shift I N B
shift I var J --> shiftindex I J
shift I lam M --> lam shift ^I M
shiftindex 0 J --> ^J
shiftindex ^I 0 --> 0
shiftindex ^I ^J --> ^shiftindex I J

if we add a lambda-reduction rule to this substitution calculus,
we get a lambda-sigma calculus.

strong normalization is obtained by having arbitrary beta-reduction:
app lam M N --> subst 0 N M

head normalization is expressible by tagging the top,
looking down for the head, and consing arguments:
Calc ::= Compute Term | Apply Term Termlist | Done
Termlist ::= nil | cons Term Termlist
Compute T --> Apply T nil
Apply app M N L --> Apply M cons N L
Apply index I L --> Done
Apply lam M nil --> Done
Apply lam M cons N L --> Apply subst N 0 M L

call-by-value is expressible, by the following rules:
Calc ::= Compute Term | Apply Calc Calc | Normal Term
Compute app M N --> Apply Compute M Compute N
Compute lam M --> Normal lam M
Apply Normal lam M Normal N --> Compute subst N 0 M

Note that some forms have no canonical meanings
in the context of call-by-value, and rules can be added or not
to detect their existence and trap into an error condition:
Compute index I --> <undefined>
Normal app M N --> <undefined>

if we only allow for lambdas as values, we get the simpler:
Calc ::= Compute Term | Apply Calc Calc | LamVal Term
Compute app M N --> Apply Compute M Compute N
Compute lam M --> LamVal M
Apply LamVal M LamVal N --> Compute subst lam N 0 M
"

(define (xsym n)
  (if (>= n 0)
      (make-symbol "_" n)
      (make-symbol "x" (- n 1))))

(define (index? n) (number? n))
(define (index->num n) n)
(define (make-index n) n)

(define (deB->alpha-n n e)
  (cond ((index? e)
	 (xsym (- n 1 (index->num e))))
	((app? e)
	 (make-app
	  (deB->alpha-n n (app-fun e))
	  (deB->alpha-n n (app-arg e))))
	((lambda? e)
;	    (format #t "dBan ~a ~a ~a\n" n e (lambda-param e))
	     (make-lambda (xsym n)
		      (deB->alpha-n (add1 n) (lambda-body e))))
	(else e)))
(define (deB->alpha e) (deB->alpha-n 0 e))

(define (deB-shift-n n e)
  (cond ((index? e)
	 (if (< e n) e (add1 e)))
	((app? e)
	 (make-app
	  (deB-shift-n n (app-fun e))
	  (deB-shift-n n (app-arg e))))
	((lambda? e)
	 (if (eqv? (lambda-param e) 0)
	     (make-lambda 0
		      (deB-shift-n (add1 n) (lambda-body e)))
	     (make-lambda (lambda-param e)
		      (deB-shift-n n (lambda-body e)))))
	(else e)))
(define (deB-shift e) (deB-shift-n 0 e))

(define (deB-subst v i e)
  (cond ((index? e)
	 (cond (((< e i) e)
		((> e i) (sub1 e))
		((= e i) v))))
	((app? e)
	 (make-app
	  (deB-subst v i (app-fun e))
	  (deB-subst v i (app-arg e))))
	((lambda? e)
	 (if (eqv? (lambda-param e) 0)
	     (make-lambda 0
		      (deB-subst (deB-shift v) (add1 i) (lambda-body e)))
	     (make-lambda (lambda-param e)
		      (deB-subst v i (lambda-body e)))))
	(else e)))

(define (symbol->deB-index x l)
  (define (i n l)
    (cond ((null? l) #f)
	  ((eq? x (car l)) n)
	  (else (i (add1 n) (cdr l)))))
  (i 0 l))
(define (shuffle-deB-index n l)
  (define (s m n l)
    (cond ((null? l) (+ m n))
	  ((zero? (car l))
	   (if (zero? n) m (s (add1 m) (sub1 n) (cdr l))))
	  (else (s (add1 m) n (cdr l)))))
  (s 0 n l))

(define (alpha->deB-l l e)
  (cond ((symbol? e)
	 (or (symbol->deB-index e l) e))
	((index? e)
	 (shuffle-deB-index e l))
	((app? e)
	 (make-app
	  (alpha->deB l (app-fun e))
	  (alpha->deB l (app-arg e))))
	((lambda? e)
	 (make-lambda 0 (alpha->deB-l (cons (lambda-param e) l)
				      (lambda-body e))))
	(else e)))
(define (alpha->deB e) (alpha->deB-l '() e))

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
"
replacing lambdas with SKI combinators (w/o subst):
Index ::= '0' | '^' Index
Term ::=  'app' Term Term | 'var' Index | 'lam' Term | 'xS' | 'xK' | 'xI'
	| 'unlambda' Term | 'unlambdan' Index Term | 'error'

Removing the lambdas from a pure lambda term (no S,K,I) is done in the
following way:

unlambda M --> unlambdan 0 M
unlambdan 0 app M N --> app unlambdan 0 M unlambdan 0 N
unlambdan 0 var I --> I
unlambdan I lam M --> unlambdan ^I M
unlambdan ^I var J -->
	if J == I
		then unlambdan I xI
		else unlambdan I app xK var unshiftindex J
unlambdan ^I app M N --> unlambdan I app app S unlambdan ^0 M unlambdan ^0 N

an additional, conditional, rule would be (using substitution as defined above)
unlambdan ^I M --> unlambdan I app K subst error I M
	*IF* M never refers variable I
It could be added as is, making the calculus non-confluent, or ignored,
or always tried first (which is tricky, but doable in a simple rewrite logic).

Accounting for the optional rule with free? this is
a non-confluent rewrite system and/or a non-determinist function
but it is *injective*, and there is a converse rewrite system!
"
; first way
(define (unlambdax x e)
  (cond ((eq? x e) 'I)
	((free? x e) (make-app 'K (unlambda-v1 e)))
	((app? e) (make-app
		       (make-app 'S (unlambdax x (app-fun e)))
		        (unlambdax x (app-arg e))))
	((lambda? e) (unlambdax x
		       (unlambdax (lambda-param e) (lambda-body e))))
	(else (make-app 'K e))))
(define (unlambda-v1 e)
  (cond	((app? e) (make-app (unlambda (app-fun e)) (unlambda (app-arg e))))
	((lambda? e) (unlambdax (lambda-param e) (lambda-body e)))
	(else e)))

; more synthetic way
(define (unlambdal l e)
  (cond
   ((lambda? e) (unlambdal (cons (lambda-param e) l) (lambda-body e)))
   ((null? l)
     (if (app? e)
	 (make-app (unlambdal l (app-fun e)) (unlambdal l (app-arg e)))
	 e))
   ((free? (car l) e) (unlambdal (cdr l) (make-app 'K e)))
   ((app? e)
     (unlambdal (cdr l)
       (make-app
	 (make-app 'S (unlambdal (list (car l)) (app-fun e)))
	 (unlambdal (list (car l)) (app-arg e)))))
   ((eq? e (car l))
     (unlambdal (cdr l) 'I))
   (else
     (unlambdal (cdr l) (make-app 'K e)))))
(define (unlambda-v2 e) (unlambdal '() e))

(define unlambda unlambda-v2)

"
Converse operation:
replacing SKI combinators (w/o subst) with lambdas
only fully works on terms obtained from lambdas
by the previous transformation (including non-determinism),
for which it retrieves the original lambda-term, up to alpha-conversion
uncombn applies in reverse topological order the rules in unlambdal...
we could use lists of gensym'd symbols, or rather, to stay purely functional,
maintain a list of symbols not already generated in a given context.
we use reverse deBruijn indices instead, which gives the same result,
but with constant- (actually, logarithmic) time (internal) integers
rather than linear-time (naive) lists [we could have logarithmic non-naive
lists, too, but my, oh my! let's stay simple].
in any case, the painful part is having to unify symbols,
which is relatively simple with indices that get shifted,
rather than alists of symbols that get renamed.

BEWARE! This function only works with terms that were actually generated
(or could have been) from the previous transformation!
The language of accepted SKI combinations is L_0 in the family of
languages L_n where:
A:L_0 (A being an atom)
I:L_0 ; K:L_1 ; S:L_2 ;
x:L_n |- x:L_n+1 ;
x:L_n |- (K x): L_n+1 ;
x:L_n+1 |- y:L_n |- (S x y):L_n

Note that (S K X) can be seamlessly used instead of I;
it will be properly recognized since it is not otherwise recognizable
in the above language family.
This family can alternatively be understood as a type system
for SKI combinations, where L_0 is the type of actual values,
that is such that L_0 == L_0 -> L_0, and L_n+1 is P->L_n,
where P is the type of all program texts (including values).
"

(define (uncombi-n n e)
  (define (uncombi n e)
;  (define (u n e)
  (cond
   ((or (eq? 'I e)
	(and (app? e)
	     (app? (app-fun e))
	     (eq? 'S (app-fun (app-fun e)))
	     (eq? 'K (app-arg (app-fun e)))))
;	  (format #t "case: I: n=~a\n" n e)
    (uncombi (add1 n) (make-index 0)))

   ((and (app? e) (eq? 'K (app-fun e)))
;	  (format #t "case: K e: n=~a e=~a\n" n e)
    (uncombi (add1 n) (deB-shift (uncombi 0 (app-arg e)))))

   ((and (app? e)
	 (app? (app-fun e))
	 (eq? 'S (app-fun (app-fun e))))
;	  (format #t "case: S M N: n=~a M=~a N=~a\n"
;		  n (app-arg (app-fun e)) (app-arg e))
    (let ((M (uncombi 0 (app-arg (app-fun e))))
	  (N (uncombi 0 (app-arg e))))
      (if (and (lambda? M)
	       (lambda? N))
	  (begin
;	  (format #t "params: ~a ~a ~a\n"
;	        (lambda-param M) (lambda-param N) 0)
	  (uncombi (add1 n)
	    (make-app (lambda-body M) (lambda-body N))))
	  (error 'uncomb n 'S M N))))

   ((zero? n)
;	  (format #t "case: n=~a e=~a\n" n e)
    (if (app? e)
	(make-app (uncombi n (app-fun e)) (uncombi n (app-arg e)))
	e))
   (else
;    (format #t "last resort: n=~a e=~a\n" n e)
    (uncombi (sub1 n) (make-lambda 0 e)))))
;    (format #t "enter uncombi: ~a ~a\n" n e)
;    (let ((res (u n e)))
;    (format #t "exit uncombi: ~a ~a ==> ~a\n" n e res)
;    res))
  (uncombi n e))
(define (uncombi e) (deB->alpha (uncombi-n 0 e)))

(define (uncomb-dbg e) (uncombl-dbg '() e))
(define (uncomb-dbg e) (uncombi e))


(define (unlambda-combn? n e)
  (define (u n e)
  (cond
   ((or (eq? 'I e)
	(and (app? e)
	     (app? (app-fun e))
	     (eq? 'S (app-fun (app-fun e)))
	     (eq? 'K (app-arg (app-fun e)))))
    #t)

   ((and (app? e) (eq? 'K (app-fun e)))
    (u (sub1 n) e))

   ((and (app? e)
	 (app? (app-fun e))
	 (eq? 'S (app-fun (app-fun e))))
    (let ((M (uncombi 0 (app-arg (app-fun e))))
	  (N (uncombi 0 (app-arg e))))
      (if (and (lambda? M)
	       (lambda? N))
	  (begin
	  (uncombi (add1 n)
	    (make-app (lambda-body M) (lambda-body N))))
	  (error 'uncomb n 'S M N))))
   ((zero? n)
    (if (app? e)
	(make-app (uncombi n (app-fun e)) (uncombi n (app-arg e)))
	e))
   (else
    (uncombi (sub1 n) (make-lambda 0 e)))))
  (u n e))

(define (unlambda-combn? n e)
  ())



;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
; normalization of terms (w/o subst)

; beta reduction
(define (do-subst x v e)  ; [v/x]e
  (cond	((free? x e) e)
	((eq? x e) v)
	((app? e) (make-app (do-subst x v (app-fun e))
			    (do-subst x v (app-arg e))))
	((lambda? e) (if (eq? x (lambda-param e)) e
			 (make-lambda (lambda-param e)
				      (do-subst x v (lambda-body e)))))
	(else e)))

(define (betareduce1 lam arg)
  (do-subst (lambda-param lam) arg (lambda-body lam)))

; generalized head reduction step
(define (headreduce1 e)
  (define (hr1r newhd rest)
    (foldapp (cons newhd rest)))
  (define (hr1b op args len)
    (cond
     ((and (eq? op 'I) (>= len 1)) (hr1r (car args) (cdr args)))
     ((and (eq? op 'K) (>= len 2)) (hr1r (car args) (cddr args)))
     ((and (eq? op 'S) (>= len 3))
      (hr1r (make-app (make-app (car args) (caddr args))
		      (make-app (cadr args) (caddr args)))
	    (cdddr args)))
     ((and (lambda? op) (>= len 1))
      (hr1r (betareduce1 op (car args)) (cdr args)))
     (else e)))
  (define (hr1 l)
    (hr1b (car l) (cdr l) (length (cdr l))))
  (hr1 (unfoldapp e)))

; head normalization (may diverge)
(define (headreduce e)
  (define (hr e f)
    (if (eq? e f) e (hr (headreduce1 e) e)))
  (hr e (not e)))


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
; enable debugging

; helper function
(define (wr0 m x)
  (map display (list m ": " x "\n\n")))
(define (wr m x)
  (map display (list m ": " x "\nunparsed: " (unparse x) "\n\n")))

(define (wr m x)
  (wr0 m (unparse x)))

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
; sample lambda terms...
(define p parse)


; Y: The beloved paradoxical combinator, compatible with applicative systems
(define eY '(lambda phi ((lambda p (p p))
			  (lambda f (phi (f f))))))
(define iY (p eY))
; B: Composition
(define eB '(lambda f (lambda g (lambda x (f (g x))))))
(define iB (p eB))

; Nrec, NO, NS: dealing with Church integers
(define eNrec '(lambda f (lambda g (lambda n (n f g)))))
(define eNO '(lambda f (lambda g f)))
(define eNS '(lambda n (lambda f (lambda g (g n)))))
(define iNrec (p eNrec))
(define iNO (p eNO))
(define iNS (p eNS))

; addition for Church integers
(define eadd `(lambda n (,eY (lambda f (,eNrec n (,eB ,eNS f))))))
(define iadd (p eadd))
; a few integer constants
(define e1 `(,eNS ,eNO))
(define ea1 `(,eadd ,e1))
(define e2 `(,ea1 ,e1))
(define i1 (p e1))
(define ia1 (p ea1))
(define i2 (p e2))

; giving integers a simpler form (requires applicative evaluation order)
(define exO 'O)
(define exS '(lambda n (S n)))
(define ex2 `(,e2 ,exS ,exO))
(define ixO (p exO))
(define ixS (p exS))
(define ix2 (p ex2))

; lambda definitions for SKI combinators
(define eI '(lambda x x))
(define eK '(lambda x (lambda y x)))
(define eS '(lambda x (lambda y (lambda z (x z (y z))))))
(define iI (p eI))
(define iK (p eK))
(define iS (p eS))

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
; current test
; !!!!!

(define iz (p `((lambda f (lambda g (lambda n (n f g)))))))
(wr 'z iz)
(define izc (unlambda iz))
(wr 'zc izc)
(define izl (uncomb-dbg izc))
(newline)
(wr 'zl izl)
(wr 'z iz)
(wr0 'a=? (alpha-equal? izl iz))

(wr 'iadd iadd)
(define iacomb1 (unlambda-v1 iadd))
(define iacomb2 (unlambda-v2 iadd))
(wr 'cmp (equal? (unparse iacomb1) (unparse iacomb2)))
(wr 'iacomb iacomb2)
(define ialam (uncomb-dbg iacomb2))
(newline)
(wr 'ialam ialam)
(wr 'iadd iadd)
(wr0 'a=? (alpha-equal? ialam iadd))



(wr 'R iNrec)
(define iRc (unlambda iNrec))
(wr 'Rc iRc)
(define iRl (uncomb-dbg iRc))
(newline)
(wr 'Rl iRl)
(wr 'R iNrec)
(wr0 'a=? (alpha-equal? iRl iNrec))

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
; a few tests

(define foo (p '(lambda 0 (lambda 0 1))))
(wr '0 foo)
(define bar (deB->alpha foo))
(wr 'a bar)
(wr0 'a=? (alpha-equal? iK bar))

(wr 'NO iNO)
(wr0 'alpha-equal? (alpha-equal? iK iNO))

(define iKcomb (unlambda iK))
(wr 'Kcomb iKcomb)
(define iKlam (uncombi iKcomb))
(wr 'Klam iKlam)
(wr0 'a=? (alpha-equal? iKlam iK))

(wr 'Y iY)
(define iYcomb (unlambda iY))
(wr 'Ycomb iYcomb)
(define iYlam (uncombi iYcomb))
(wr 'Ylam iYlam)
(wr0 'a=? (alpha-equal? iYlam iY))

(wr 'i2 i2)
(define i2comb (unlambda i2))
(wr 'i2comb i2comb)
(define i2lam (uncomb-dbg i2comb))
(wr 'i2lam i2lam)
(wr0 'a=? (alpha-equal? i2lam i2))

;(define foo (parse (read)))
;(define foo '(K S S S S))
(define foo iadd)

(wr 'foo foo)
(define bar (unlambda-v1 foo))
(wr 'unlambda-v1 bar)
(define baz (unlambda-v2 foo))
(wr 'unlambda-v2 baz)
(wr 'cmp (equal? (unparse bar) (unparse baz)))
(wr 'hr1 (headreduce1 foo))
(wr 'uhr1 (headreduce1 bar))
(define hfoo (headreduce foo))
(wr 'hr hfoo)
(define hbar (headreduce bar))
(wr 'uhr hbar)
(wr 'hru (unlambda hfoo))

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------

; ross@soi.city.ac.uk on comp.lang.functional reports
; that there are fixed point combinations of SKI-length 9:
; S I (S (K (S I))) (S I I)
; or
; S I V W
; where	V = S (K (S I))
;	W = S I I
; Other fixed points in 9 combinators are
; W (S V I)
; S (S I) (S V) I
; S S I (S V) I
; W (S I V)
; S (S I) (S I) V
; S S I (S I) V
; S (S S) S I V

; john tromp has a nice page on a universal combinator...
; http://www.cwi.nl/~tromp

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
; THE END
;(exit)

;;; Pattern matching

; V: valeurs
; P: programmes sources; sous-ensemble de V
;
'`(eval ,P ,V)
'`(apply ,V ,V)

;;; 

'(==> `(eval ,P ,V) )

;;; EOF
